package org.greenstone.gsdl3.core;

import java.io.File ;
import java.io.FileReader;

import org.apache.xerces.parsers.DOMParser;
import org.greenstone.gsdl3.util.*;
import org.w3c.dom.*;
import org.xml.sax.InputSource;

import java.util.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;


public class Skin {
	
	public File rootDirectory ;
	public Document config ;
	private Receptionist receptionist ;
	
	protected DOMParser parser = null;
	TransformerFactory transformerFactory=null;
	
	private HashMap<String, String> xsltPagesByAction ; 
	
	public Skin(Page page, Receptionist receptionist) throws Exception{
		
		this.receptionist = receptionist ;
		this.transformerFactory = org.apache.xalan.processor.TransformerFactoryImpl.newInstance();
		transformerFactory.setURIResolver(new MyUriResolver()) ;
		
		this.parser = new DOMParser();
	    this.parser.setFeature("http://xml.org/sax/features/validation", false);
	    // don't try and load external DTD - no need if we are not validating, and may cause connection errors if a proxy is not set up.
	    this.parser.setFeature("http://apache.org/xml/features/nonvalidating/load-external-dtd", false);
	    // a performance test showed that having this on lead to increased 
	    // memory use for small-medium docs, and not much gain for large 
	    // docs.
	    // http://www.sosnoski.com/opensrc/xmlbench/conclusions.html
	    this.parser.setFeature("http://apache.org/xml/features/dom/defer-node-expansion", false);
		
		
		String siteHome = GSFile.siteHome(GlobalProperties.getGSDL3Home(), page.getSite()) ;
		Document collectUi = page.getCollectUi() ;
		
		if (collectUi != null) {
			//System.out.println("looking for skin defined at collect level") ;
			
			String collectHome = GSFile.collectionBaseDir(siteHome, page.getCollection()) ;
			Element xmlSkin = (Element) GSXML.getChildByTagName(collectUi.getFirstChild(), "skin") ;
			if (xmlSkin != null) {
				//System.out.println("skin defined at collect level") ;
				
				boolean inherit = Boolean.parseBoolean(xmlSkin.getAttribute("inherit")) ;
				
				if (inherit != true) {
					boolean local = Boolean.parseBoolean(xmlSkin.getAttribute("local")) ;
					if (local == true) {
						rootDirectory = new File(collectHome + File.separatorChar + "ui") ;
					} else {
						String name = xmlSkin.getAttribute("name") ;
						rootDirectory = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "skins" + File.separatorChar + File.separatorChar + name) ;
					}
				}				
			}
		}
		
		Document siteUi = page.getSiteUi() ;
		
		if (rootDirectory == null && siteUi != null) {
			//System.out.println("lookding for skin defined at site level") ;
			
			Element xmlSkin = (Element) GSXML.getChildByTagName(siteUi.getFirstChild(), "skin") ;
			if (xmlSkin != null) {
				//System.out.println("skin defined at site level") ;
				boolean local = Boolean.parseBoolean(xmlSkin.getAttribute("local")) ;
				if (local == true) {
					rootDirectory = new File(siteHome + File.separatorChar + "ui") ;
				} else {
					String name = xmlSkin.getAttribute("name") ;
					rootDirectory = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "skins" + File.separatorChar + name) ;
				}
			}				
		}
		
		if(rootDirectory == null) {
			rootDirectory = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "skins" + File.separatorChar + "default") ;
		}

		
		File configFile = new File(rootDirectory.getAbsolutePath() + File.separatorChar + "skin.xml") ;
		config = receptionist.converter.getDOM(configFile, "utf-8");
		
		
		// store which xslt pages are responsible for which actions
		
		xsltPagesByAction = new HashMap() ;
		
		NodeList actions = config.getElementsByTagName("action") ;
		
		for (int ai=0 , an=actions.getLength() ; ai<an ; ai++){
			Element action = (Element) actions.item(ai) ;
			String a = action.getAttribute("name") ;
			
			String xsl = action.getAttribute("xslt") ;
			if (!xsl.equals("")) 
				xsltPagesByAction.put(a, xsl) ;
			
			NodeList subactions = action.getElementsByTagName("subaction") ;
			
			for (int si=0 , sn=subactions.getLength() ; si<sn ; si++){
				Element subaction = (Element) subactions.item(si) ;
				
				String sa = subaction.getAttribute("name") ;
				xsl = subaction.getAttribute("xslt") ;
				xsltPagesByAction.put(a + "_" + sa, xsl) ;
			}
		}
	}
	
	private Document getXSLTDoc(String action, String subaction) throws Exception {
		//System.out.println("getting xslt for " + action + ", " + subaction) ;
		
		String name = xsltPagesByAction.get(action + "_" + subaction) ;
		
		if (name == null) 
			name = xsltPagesByAction.get(action) ;
		
		File xslt_file = new File(rootDirectory.getAbsolutePath() + File.separatorChar + "xsl" + File.separatorChar + name) ;
		//System.out.println("Skinning page using: " + xslt_file) ;
		
		//if (!xslt_file.canRead()) 
		//	xslt_file = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "xslt" + File.separatorChar + "error.xsl") ;
		
		FileReader reader = new FileReader(xslt_file);
	    InputSource xml_source = new InputSource(reader);
	    this.parser.parse(xml_source);
	    Document doc = this.parser.getDocument();
		
		return doc ;
	}
	
	private Document getPreprocessDoc() throws Exception {
	                                                      
		File xslt_file = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "xslt" + File.separatorChar + "preProcess.xsl") ;
		
		FileReader reader = new FileReader(xslt_file);
	    InputSource xml_source = new InputSource(reader);
	    this.parser.parse(xml_source);
	    Document doc = this.parser.getDocument();
	    
	    return doc ;		
	}
	
	private Document getLibraryDoc() throws Exception {
        
		File xslt_file = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "xslt" + File.separatorChar + "library.xsl") ;
		
		FileReader reader = new FileReader(xslt_file);
	    InputSource xml_source = new InputSource(reader);
	    this.parser.parse(xml_source);
	    Document doc = this.parser.getDocument();
	    
	    return doc ;		
	}
	
	
	public Element transformPage(Page page) throws Exception{
		
		Element p = page.getPage() ;
		Element pr = page.getPageResponse() ;
		Element transformedPage = null ;
		
		Element blah = receptionist.doc.createElement("Skin") ;
		blah.setAttribute("skinLocation", rootDirectory.getAbsolutePath()) ;
		pr.appendChild(blah) ;
		
		Document sourceXml ;
		try {
			sourceXml = receptionist.converter.newDOM();
			sourceXml.appendChild(sourceXml.importNode(p, true));
		} catch (Exception e) {
			System.out.println("error loading source data") ;
			e.printStackTrace() ;
			return constructErrorPage(e) ;
		}
		
		Document skinXsl ;
		try {
			skinXsl = getXSLTDoc(page.getAction(), page.getSubaction()) ;
		} catch (Exception e) {
			System.out.println("error loading skin xslt") ;
			e.printStackTrace() ;
			return constructErrorPage(e) ;
		}
		
		Document preprocessingXsl  ;
		try {
			preprocessingXsl = getPreprocessDoc() ;
		} catch (Exception e) {
			System.out.println("error loading preprocessing xslt") ;
			e.printStackTrace() ;
			return constructErrorPage(e) ;
		}
		
		Document libraryXsl ;
		try {
			libraryXsl = getLibraryDoc() ;
		} catch (Exception e) {
			System.out.println("error loading preprocessing xslt") ;
			e.printStackTrace() ;
			return constructErrorPage(e) ;
		}
		
		// combine skin file and library variables/templates into one document. 
		//   we dont just use xsl:import because the preprocessing stage needs 
		//   to know what's available in the library.
		
		Document skinAndLibraryXsl ;
		Document processedXsl = receptionist.converter.newDOM();
		try {
			
			skinAndLibraryXsl = receptionist.converter.newDOM();
			Element root = skinAndLibraryXsl.createElement("skinAndLibraryXsl") ;
			skinAndLibraryXsl.appendChild(root) ;
						
			Element s = skinAndLibraryXsl.createElement("skinXsl") ;
			s.appendChild(skinAndLibraryXsl.importNode(skinXsl.getDocumentElement(), true)) ;
			root.appendChild(s) ;
			
			Element l = skinAndLibraryXsl.createElement("libraryXsl") ;
			l.appendChild(skinAndLibraryXsl.importNode(libraryXsl.getDocumentElement(), true)) ;
			root.appendChild(l) ;
			
			
			System.out.println("Pre - processing") ;
			//pre-process the skin style sheet
			Transformer preProcessor = transformerFactory.newTransformer(new DOMSource(preprocessingXsl));
			DOMResult result = new DOMResult();
			result.setNode(processedXsl) ;
			preProcessor.transform(new DOMSource(skinAndLibraryXsl), result);
		} catch (TransformerException e) {
			return constructErrorPage(e) ;
		}  catch (Exception e) {
			System.out.println("error preprocessing the skin xslt") ;
			e.printStackTrace() ;
			return constructErrorPage(e) ;
		}
		
		//return sourceXml.getDocumentElement() ;
		
		Document finalPage = receptionist.converter.newDOM();
		//transform source doc
		try {
			Transformer transformer = transformerFactory.newTransformer(new DOMSource(processedXsl));
			DOMResult result = new DOMResult();
			result.setNode(finalPage) ;
			transformer.transform(new DOMSource(sourceXml), result);
		} catch (TransformerException e) {
			return constructErrorPage(e) ;
		} catch (Exception e) {
		
			System.out.println("error transforming page") ;
			e.printStackTrace() ;
			
			//return processedXsl.getDocumentElement() ;
			return constructErrorPage(e) ;
		}
		
		return finalPage.getDocumentElement() ;
	}
	
	private Element constructErrorPage(TransformerException exception) {
		Element page = receptionist.doc.createElement("page") ;
		
		Element header = receptionist.doc.createElement("h1") ;
		header.setTextContent("Error") ;
		page.appendChild(header) ;
		
		Element prompt = receptionist.doc.createElement("p") ;
		prompt.setTextContent("The following exception occured: ") ;
		page.appendChild((prompt)) ;
		
		Element errorXml = receptionist.doc.createElement("code") ;
		errorXml.setTextContent(exception.getMessageAndLocation()) ;
		page.appendChild(errorXml) ;
		
		return page ;
		
	}
	
	private Element constructErrorPage(Exception exception) {
		Element page = receptionist.doc.createElement("page") ;
		
		Element header = receptionist.doc.createElement("h1") ;
		header.setTextContent("Error") ;
		page.appendChild(header) ;
		
		Element prompt = receptionist.doc.createElement("p") ;
		prompt.setTextContent("The following exception occured: ") ;
		page.appendChild((prompt)) ;
		
		Element errorXml = receptionist.doc.createElement("code") ;
		errorXml.setTextContent(exception.toString()) ;
		page.appendChild(errorXml) ;
		
		
		return page ;
	}
	

	private Element constructErrorPage(Element source, Document style, Exception exception) {
		Element page = receptionist.doc.createElement("page") ;
		
		Element header = receptionist.doc.createElement("h1") ;
		header.setTextContent("Error") ;
		page.appendChild(header) ;
		
		Element prompt = receptionist.doc.createElement("p") ;
		prompt.setTextContent("The following exception occured: ") ;
		page.appendChild((prompt)) ;
		
		Element errorXml = receptionist.doc.createElement("code") ;
		errorXml.setTextContent(exception.getMessage()) ;
		page.appendChild(errorXml) ;
		
		/*
		Element stackTrace = receptionist.doc.createElement("ul") ;
		
		StackTraceElement[] st = exception.getStackTrace() ;
		for (int i=0 ; i< st.length ; i++) {
			Element ste = receptionist.doc.createElement("li") ;
			ste.setTextContent(st[i].toString()) ;
			stackTrace.appendChild(ste) ;
		}
		errorMessage.appendChild(stackTrace) ;
		*/
		
		Element sourceHeader = receptionist.doc.createElement("h2") ;
		sourceHeader.setTextContent("Source page:") ;
		page.appendChild(sourceHeader) ;
		
		Element sourceXml = receptionist.doc.createElement("pre") ;
		sourceXml.setTextContent(receptionist.converter.getPrettyString(source)) ;
		page.appendChild(sourceXml) ;
		
		Element styleHeader = receptionist.doc.createElement("h2") ;
		styleHeader.setTextContent("Style page:") ;
		page.appendChild(styleHeader) ;
		
		Element styleXml = receptionist.doc.createElement("pre") ;
		styleXml.setTextContent(receptionist.converter.getPrettyString(style)) ;
		page.appendChild(styleXml) ;
		
		
		return (Element)page ;
	}
	
	private class MyUriResolver implements URIResolver {
		
		public Source resolve(String href, String base) {
			
			System.out.println("resolving href='" + href + "', base='" + base + "'") ;
						
			// check in the skin directory first
			
			File file = new File(rootDirectory.getAbsolutePath() + File.separatorChar + "xslt" + File.separatorChar + href) ;
			
			// then check in the xslt library directory
			if (!file.canRead()) 
				file = new File(GlobalProperties.getGSDL3Home() + File.separatorChar + "ui" + File.separatorChar + "xslt" + File.separatorChar + href) ;
			
			if (file.canRead()) {
				Source source = new StreamSource(file) ;
				return source ;
			} else
				return null ;
		}
		
	}
	
}	

